import logging
import re
from typing import Any, Optional, Union

import pandas as pd
from click import File

log = logging.getLogger("ftm_columnstore")


def read_csv(
    infile: Union[File, str],
    do_raise: Optional[bool] = True,
    fillna: Optional[Any] = "",
    delimiter: Optional[str] = ",",
    **kwargs,
) -> pd.DataFrame:
    read_kwargs = {**{"on_bad_lines": "warn"}, **kwargs}
    df = None
    if isinstance(infile, str):
        fname = infile
    else:
        fname = infile.name

    try:
        if fname.endswith(".gz"):
            df = pd.read_csv(fname, compression="gzip", **read_kwargs)
        else:
            df = pd.read_csv(infile, **read_kwargs)
    except Exception as e:
        handle_error(log, e, do_raise, fpath=infile.name)
    if df is not None:
        return df.fillna(fillna)


def handle_error(
    logger, e: Union[Exception, str], do_raise: bool, dataset: Optional[str] = None
):
    if isinstance(e, str):
        e = Exception(e)
    if do_raise:
        raise e
    msg = ""
    if dataset is not None:
        msg = f"[{dataset}] "
    logger.error(msg + f"{e.__class__.__name__}: `{e}`")


NUMERIC_US = re.compile(r"^-?\d+(?:,\d{3})*(?:\.\d+)?$")
NUMERIC_DE = re.compile(r"^-?\d+(?:\.\d{3})*(?:,\d+)?$")


def to_numeric(value: str):
    value = str(value).strip()
    try:
        value = float(value)
        if int(value) == value:
            return int(value)
        return value
    except ValueError:
        if re.match(NUMERIC_US, value):
            return to_numeric(value.replace(",", ""))
        if re.match(NUMERIC_DE, value):
            return to_numeric(value.replace(".", "").replace(",", "."))



# https://github.com/opensanctions/opensanctions/blob/main/opensanctions/core/statements.py
from datetime import datetime
from functools import lru_cache
from hashlib import sha1
from typing import Iterator, Optional, TypedDict

import fingerprints.generate
from followthemoney import model
from followthemoney.util import make_entity_id


@lru_cache(maxsize=1024 * 1000)  # 1GB
def _get_fingerprint_id(fingerprint: str) -> str:
    return make_entity_id(fingerprint)


class Statement(TypedDict):
    """A single statement about a property relevant to an entity.
    For example, this could be useddocker to say: "In dataset A, entity X has the
    property `name` set to 'John Smith'. I first observed this at K, and last
    saw it at L."
    Null property values are not supported. This might need to change if we
    want to support making property-less entities.
    """

    id: str
    dataset: str
    entity_id: str
    schema: str
    prop: str
    prop_type: str
    value: str
    value_num: Optional[float] = None
    last_seen: datetime


class FingerprintStatement(TypedDict):
    """A statement describing a fingerprint for an entity"""

    id: str
    dataset: str
    entity_id: str
    schema: str
    fingerprint: str
    fingerprint_id: str


def stmt_key(dataset: str, entity_id: str, prop: str, value: str) -> str:
    """Hash the key properties of a statement record to make a unique ID."""
    key = f"{dataset}.{entity_id}.{prop}.{value}"
    return sha1(key.encode("utf-8")).hexdigest()


def statements_from_entity(entity: dict, dataset: str) -> Iterator[Statement]:
    entity = model.get_proxy(entity)
    if entity.id is None or entity.schema is None:
        return []
    entity_id = entity.id.rsplit(".", 1)[0]
    for prop, value in entity.itervalues():
        if value:
            value_num = None
            if prop.type.name == "number":
                value_num = value
            elif prop.type.name == "entity":
                value = value.rsplit(".", 1)[0]
            stmt: Statement = {
                "id": stmt_key(dataset, entity_id, prop.name, value),
                "dataset": dataset,
                "entity_id": entity_id,
                "schema": entity.schema.name,
                "prop": prop.name,
                "prop_type": prop.type.name,
                "value": value,
                "value_num": value_num,
                "last_seen": datetime.now().isoformat(),
            }
            yield stmt


def _should_fingerprint(entity):
    if entity.id is None or entity.schema is None:
        return False
    if entity.schema.is_a("Mention"):
        return True
    return entity.schema.is_a("Thing")


def fingerprints_from_entity(
    entity: dict, dataset: str
) -> Iterator[FingerprintStatement]:
    entity = model.get_proxy(entity)
    if not _should_fingerprint(entity):
        return []
    entity_id = entity.id.rsplit(".", 1)[0]
    for prop, value in entity.itervalues():
        if value:
            if prop.type.name == "name":
                fingerprint = fingerprints.generate(value)
                fingerprint_id = make_entity_id(fingerprint)
                stmt: FingerprintStatement = {
                    "id": stmt_key(dataset, entity_id, prop.name, value),
                    "dataset": dataset,
                    "entity_id": entity_id,
                    "schema": entity.schema.name,
                    "prop": prop.name,
                    "fingerprint": fingerprint,
                    "fingerprint_id": fingerprint_id,
                }
                yield stmt


COLUMNS = tuple(Statement.__annotations__.keys())
COLUMNS_FPX = tuple(FingerprintStatement.__annotations__.keys())



from functools import lru_cache
from typing import Any, Iterable, Iterator, Optional

import pandas as pd
from clickhouse_driver import Client, errors

from . import settings

from .util import to_numeric


def table_exists(e: Exception, table: str) -> bool:
    return f"Table default.{table} already exists" in str(e)


class ClickhouseDriver:
    def __init__(
        self,
        uri: Optional[str] = settings.DATABASE_URI,
        table: Optional[str] = settings.DATABASE_TABLE,
    ):
        self.table = table
        self.table_fpx = f"{table}_fpx"
        self.uri = uri
        self.ensure_table()

    def __str__(self):
        return self.uri

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self})>"

    def init(self, recreate: Optional[bool] = False, exists_ok: Optional[bool] = False):
        if recreate:
            self.dangerous_drop()
        for stmt in self.create_statements:
            try:
                self.execute(stmt)
            except Exception as e:
                if exists_ok and (
                    table_exists(e, self.table) or table_exists(e, self.table_fpx)
                ):
                    pass
                else:
                    raise e

    def dangerous_drop(self):
        for stmt in self.drop_statements:
            self.execute(stmt)

    def ensure_table(self):
        try:
            self.init(recreate=False)
        except errors.ServerException as e:
            if table_exists(e, self.table) or table_exists(e, self.table_fpx):
                pass
            else:
                raise e

    def insert(self, df: pd.DataFrame, table: Optional[str] = None) -> int:
        # https://clickhouse-driver.readthedocs.io/en/latest/features.html#numpy-pandas-support
        table = table or self.table
        if "value_num" in df:
            df["value_num"] = df["value_num"].map(to_numeric)
        df = df.applymap(lambda x: None if x == "" or pd.isna(x) else x)
        with self.get_connection() as conn:
            res = conn.insert_dataframe("INSERT INTO %s VALUES" % table, df)
        return res

    def query(self, query: Any, *args, **kwargs) -> Iterator[Any]:
        query = str(query)
        conn = self.get_connection()
        kwargs = {**{"settings": {"max_block_size": 100000}}, **kwargs}
        return conn.execute_iter(query, *args, **kwargs)
        # with self.get_connection() as conn:
        #     res = conn.execute_iter(query, settings={"max_block_size": 100000})
        # return res

    def query_dataframe(self, query: Any) -> pd.DataFrame:
        query = str(query)
        with self.get_connection() as conn:
            return conn.query_dataframe(query)

    def get_connection(
        self,
        uri: Optional[str] = settings.DATABASE_URI,
    ):
        host, *port = uri.split(":")
        if not port:
            port = [9000]
        return Client(settings={"use_numpy": True}, host=host, port=port[0])

    def execute(self, *args, **kwargs):
        with self.get_connection() as conn:
            res = conn.execute(*args, **kwargs)
        return res

    def execute_iter(self, *args, **kwargs):
        return self.query(*args, **kwargs)

    @property
    def create_statements(self) -> Iterable[str]:
        UNIQUE = ("dataset", "entity_id", "schema", "prop", "origin", "value")
        CREATE_TABLE = """
        CREATE TABLE {table}
        (
            `id`                      FixedString(40) NOT NULL,
            `dataset`                 String NOT NULL,
            `entity_id`               String NOT NULL,
            `schema`                  LowCardinality(String) NOT NULL,
            `origin`                  String NOT NULL,
            `prop`                    LowCardinality(String) NOT NULL,
            `prop_type`               LowCardinality(String) NOT NULL,
            `value`                   String NOT NULL,
            `value_num`               Float NULL,
            `last_seen`               DateTime NOT NULL
        ) ENGINE = ReplacingMergeTree(last_seen)
        PRIMARY KEY ({primary_key})
        ORDER BY ({order_by})
        """
        CREATE_TABLE_FPX = """
        CREATE TABLE {table}
        (
            `id`                      FixedString(40) NOT NULL,
            `dataset`                 String NOT NULL,
            `entity_id`               String NOT NULL,
            `schema`                  LowCardinality(String) NOT NULL,
            `prop`                    LowCardinality(String) NOT NULL,
            `fingerprint`             String NOT NULL,
            `fingerprint_id`          FixedString(40) NOT NULL,
            INDEX fp_ix (fingerprint) TYPE ngrambf_v1(3, 256, 2, 0) GRANULARITY 4
        ) ENGINE = ReplacingMergeTree()
        PRIMARY KEY ({primary_key})
        ORDER BY ({order_by})
        """

        # implicit enum types
        primary_key = ", ".join(UNIQUE[:4])
        order_by = ", ".join(UNIQUE)
        create_table = CREATE_TABLE.format(
            table=self.table,
            primary_key=primary_key,
            order_by=order_by,
        )
        create_table_fpx = CREATE_TABLE_FPX.format(
            table=self.table_fpx,
            primary_key="fingerprint_id,schema,dataset",
            order_by="fingerprint_id,schema,dataset",
        )
        projection_values = f"""
        ALTER TABLE {self.table} ADD PROJECTION {self.table}_values (
            SELECT *
            ORDER BY value,prop,dataset,schema
        )
        """
        projection_entities = f"""
        ALTER TABLE {self.table} ADD PROJECTION {self.table}_entities (
            SELECT dataset,entity_id,schema,prop,groupUniqArray(value) as values
            GROUP BY dataset,entity_id,schema,prop
        )
        """
        return (
            create_table,
            create_table_fpx,
            projection_values,
            projection_entities,
        )

    @property
    def drop_statements(self) -> str:
        return (
            f"DROP TABLE IF EXISTS {self.table}",
            f"DROP TABLE IF EXISTS {self.table_fpx}",
        )


@lru_cache(128)
def get_driver(
    uri: Optional[str] = None,
    table: Optional[str] = None,
) -> ClickhouseDriver:

    # this allows overwriting settings during runtime (aka tests)
    uri = uri or settings.DATABASE_URI
    table = table or settings.DATABASE_TABLE
    return ClickhouseDriver(uri, table)



import logging
from functools import lru_cache
from typing import Iterator, Optional, Union

import pandas as pd
from followthemoney.proxy import E

from . import settings
from .driver import ClickhouseDriver, get_driver
from .query import EntityQuery, Query
from .statements import fingerprints_from_entity, statements_from_entity
from .util import handle_error

log = logging.getLogger(__name__)


class BulkWriter:  # FIXME make seperated fingerprints writer?
    def __init__(
        self,
        dataset: "Dataset",
        with_fingerprints: Optional[bool] = False,
        origin: Optional[str] = None,
        size: Optional[int] = settings.BULK_WRITE_SIZE,
        ignore_errors: Optional[bool] = False,
    ):
        self.dataset = dataset
        self.with_fingerprints = with_fingerprints
        self.origin = origin or dataset.origin
        self.buffer = []
        self.buffer_fpx = []
        self.size = size
        self.ignore_errors = ignore_errors
        self.entities = 0
        self.statements = 0
        self.statements_fpx = 0

    def put(self, entity: Union[dict, E]):
        if hasattr(entity, "to_dict"):
            entity = entity.to_dict()
        else:
            entity = dict(entity)
        self.entities += 1
        for statement in statements_from_entity(entity, self.dataset.name):
            self.buffer.append(statement)
            self.statements += 1
        if self.with_fingerprints:
            for statement in fingerprints_from_entity(entity, self.dataset.name):
                self.buffer_fpx.append(statement)
                self.statements_fpx += 1
        if self.entities % self.size == 0:
            self.flush()

    def flush(self):
        if not len(self.buffer):
            return
        try:
            df = pd.DataFrame(self.buffer)
            df["origin"] = self.origin
            res = self.dataset.driver.insert(df)
            log.info(
                f"[{self.dataset}] Write: {self.entities} entities with {self.statements} statements."
            )
            if self.with_fingerprints:
                df = pd.DataFrame(self.buffer_fpx)
                df["origin"] = self.origin
                self.dataset.driver.insert(df, table=self.dataset.driver.table_fpx)
                log.info(
                    f"[{self.dataset}] Write: {self.statements_fpx} fingerprint statements."
                )
            self.buffer = []
            self.buffer_fpx = []
            return res
        except Exception as e:
            handle_error(log, e, not self.ignore_errors)


class Dataset:
    def __init__(
        self,
        name: str,
        origin: Optional[str] = None,
        driver: Optional[ClickhouseDriver] = None,
        ignore_errors: Optional[bool] = False,
    ):
        self.name = name
        self.origin = origin
        self.driver = driver or get_driver()
        self.ignore_errors = ignore_errors
        self.query = EntityQuery(driver=driver).where(dataset=name, origin=origin)

    def drop(self):
        log.info("Dropping ftm-store: %s" % self.name)
        where = self.query.where_part
        stmt = f"ALTER TABLE {self.driver.table} DELETE {where}"
        try:
            return self.driver.execute(stmt)
        except Exception as e:
            handle_error(
                log,
                e,
                not self.ignore_errors,
                dataset=self.name,
            )

    def delete(self, entity_id: Optional[str] = None, origin: Optional[str] = None):
        if entity_id or origin:
            where = Query(driver=self.driver).where(entity_id=entity_id, origin=origin)
            stmt = f"ALTER TABLE {self.driver.table} DELETE {where}"
            return self.driver.execute(stmt)
        return self.drop()

    def put(self, entity, origin: Optional[str] = None):
        bulk = self.bulk(origin=origin or self.origin)
        bulk.put(entity)
        return bulk.flush()

    def bulk(
        self, origin: Optional[str] = None, with_fingerprints: Optional[bool] = False
    ) -> "BulkWriter":
        return BulkWriter(
            self,
            with_fingerprints=with_fingerprints,
            origin=origin or self.origin,
            ignore_errors=self.ignore_errors,
        )

    def iterate(
        self, entity_id: Optional[str] = None, chunksize: Optional[int] = 1000
    ) -> Iterator[E]:
        q = self.query
        if entity_id is not None:
            q = q.where(entity_id=entity_id)
        return q.iterate(chunksize=chunksize)

    def statements(self, origin: Optional[str] = None) -> Iterator[tuple]:
        q = (
            Query(driver=self.driver)
            .select("DISTINCT ON (id) *")
            .where(dataset=self.name, origin=self.origin)
        )
        return q.iterate()

    def get(self, entity_id: str):
        for entity in self.iterate(entity_id=entity_id):
            return entity

    def expand(self, entity: E, levels: Optional[int] = 1) -> Iterator[E]:
        """
        find connected entities in both directions
        """

        def _expand(entity: E, levels: int):
            # outgoing
            yield from self.resolve(entity, levels)
            # incoming
            query = EntityQuery(driver=self.driver).where(
                prop_type="entity", value=entity.id
            )
            for entity in query:
                yield entity
                if levels - 1 > 0:
                    yield from self.resolve(entity, levels - 1)
                    yield from _expand(entity, levels - 1)

        # uniq
        entities = set()
        for e in _expand(entity, levels):
            entities.add(e)
        yield from entities

    def resolve(self, entity: E, levels: Optional[int] = 1) -> Iterator[E]:
        """
        resolve entity prop values to actual entites
        """

        def _resolve(entity: E, levels: int):
            for prop, value in entity.itervalues():
                if prop.type.name == "entity":
                    expanded_entity = self.get(value)
                    yield expanded_entity
                    if levels - 1 > 0:
                        yield from _resolve(expanded_entity, levels - 1)

        # uniq
        entities = set()
        for e in _resolve(entity, levels):
            entities.add(e)
        yield from entities

    def __iter__(self):
        return self.iterate()

    def __len__(self):
        return len(self.query)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Dataset(%r, %r)>" % (self.driver, self.name)


@lru_cache(maxsize=128)
def get_dataset(
    name: str,
    driver: Optional[ClickhouseDriver] = None,
    origin: Optional[str] = None,
    ignore_errors: Optional[bool] = False,
):
    driver = driver or get_driver()
    return Dataset(name, origin, driver, ignore_errors)



from tests.util import ClickhouseTestCase

from ftm_columnstore.exceptions import InvalidQuery
from ftm_columnstore.query import Query, EntityQuery


class QueryTestCase(ClickhouseTestCase):
    maxDiff = None

    def test_query(self):
        q = Query()
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test")

        q = Query().select()
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test")

        q = Query("ftm_test")
        self.assertEqual(str(q), "SELECT * FROM ftm_test")

        # accept inner queries
        inner = Query().where(dataset="foo")
        q = Query(inner).select("count(distinct prop)")
        self.assertEqual(
            str(q),
            "SELECT count(distinct prop) FROM (SELECT * FROM ftm_columnstore_test WHERE dataset = 'foo')",
        )

        q = Query().select("entity_id", "prop")
        self.assertEqual(str(q), "SELECT entity_id, prop FROM ftm_columnstore_test")

        q = Query().select().where(country="de")
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'country' AND value = 'de')",
        )

        q = Query().select().where(country="de", date=2019)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'country' AND value = 'de') OR (prop = 'date' AND value = '2019')",
        )

        q = Query().group_by("prop", "entity_id")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test GROUP BY entity_id, prop"
        )

        q = Query().order_by("value")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test ORDER BY value ASC"
        )

        q = Query().order_by("prop", "value", ascending=False)
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test ORDER BY prop, value DESC"
        )

    def test_query_where_operators(self):
        """
        .where(prop=value) turns into (prop = '$prop' AND value = '$value')
        or `value_num` for numeric properties based on ftm_columnstore_test schema
        except meta fields (dataset, entity_id, schema, origin)
        """
        q = Query().where(dataset="foo")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test WHERE dataset = 'foo'"
        )

        q = Query().where(schema="Person")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test WHERE schema = 'Person'"
        )

        q = Query().where(entity_id="foo")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test WHERE entity_id = 'foo'"
        )

        q = Query().where(origin="foo")
        self.assertEqual(
            str(q), "SELECT * FROM ftm_columnstore_test WHERE origin = 'foo'"
        )

        q = Query().where(schema="Payment", amount__gt=0)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE schema = 'Payment' AND (prop = 'amount' AND value_num > '0')",
        )

        # multiple ftm_columnstore_test prop lookups are combined with "OR" because of statement structure
        q = Query().where(name="foo", summary="bar")
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value = 'foo') OR (prop = 'summary' AND value = 'bar')",
        )

        q = Query().where(name__like="nestle")
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value LIKE 'nestle')",
        )

        q = Query().where(name__ilike="nestle")
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value ILIKE 'nestle')",
        )

        q = Query().where(amount__gt=10)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'amount' AND value_num > '10')",
        )

        q = Query().where(amount__gte=10)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'amount' AND value_num >= '10')",
        )

        q = Query().where(amount__lt=10)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'amount' AND value_num < '10')",
        )

        q = Query().where(amount__lte=10)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'amount' AND value_num <= '10')",
        )

        q = Query().where(name__in=("alice", "lisa"))
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value IN ('alice', 'lisa'))",
        )

        # IN will accept subqueries
        inner = Query().select("DISTINCT entity_id").where(name="alice")
        q = Query().where(entity_id__in=inner)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE entity_id IN (SELECT DISTINCT entity_id FROM ftm_columnstore_test WHERE (prop = 'name' AND value = 'alice'))",
        )

    def test_query_slice(self):
        q = Query()[:100]
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test LIMIT 0, 100")

        q = Query()[100:200]
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test LIMIT 100, 100")

        q = Query()[100:]
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test OFFSET 100")

        q = Query()[17]
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test LIMIT 17, 1")

    def test_query_having(self):
        q = (
            Query()
            .select("entity_id", "sum(value_num) as amount_sum")
            .where(schema="Payment", amount__gt=0)
            .group_by("entity_id")
            .having(amount_sum__gte=100)
        )
        self.assertEqual(
            str(q),
            "SELECT entity_id, sum(value_num) as amount_sum FROM ftm_columnstore_test WHERE schema = 'Payment' AND (prop = 'amount' AND value_num > '0') GROUP BY entity_id HAVING amount_sum >= '100'",
        )

        # no having if no group by
        q = Query().having(foo="bar")
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test")
        self.assertEqual(
            str(q.group_by("foo")),
            "SELECT * FROM ftm_columnstore_test GROUP BY foo HAVING foo = 'bar'",
        )

    def test_query_correct_chain(self):
        q = (
            Query()
            .select("a")
            .where(name="bar")
            .select("b", "c")
            .where(amount=1, summary="f")
        )
        self.assertEqual(
            str(q),
            "SELECT a, b, c FROM ftm_columnstore_test WHERE (prop = 'amount' AND value_num = '1') OR (prop = 'name' AND value = 'bar') OR (prop = 'summary' AND value = 'f')",
        )

        # group by should be combined
        q = Query().group_by("a").group_by("b")
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test GROUP BY a, b")

        # order by should be overwritten!
        q = Query().order_by("a").order_by("b")
        self.assertEqual(str(q), "SELECT * FROM ftm_columnstore_test ORDER BY b ASC")

    def test_query_lookup_null(self):
        q = Query().where(name__null=True)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value IS NULL)",
        )
        q = Query().where(name__null=False)
        self.assertEqual(
            str(q),
            "SELECT * FROM ftm_columnstore_test WHERE (prop = 'name' AND value IS NOT NULL)",
        )

    def test_query_invalid(self):
        with self.assertRaisesRegex(InvalidQuery, "must not be negative"):
            q = Query()[-1]
            str(q)

        with self.assertRaisesRegex(InvalidQuery, "must not be negative"):
            q = Query()[100:50]
            str(q)

        with self.assertRaisesRegex(InvalidQuery, "steps not allowed"):
            q = Query()[100:50:2]
            str(q)

        with self.assertRaisesRegex(InvalidQuery, "Invalid operator"):
            q = Query().where(name__invalid_op=0)
            str(q)

        with self.assertRaisesRegex(InvalidQuery, "Invalid operator"):
            q = Query().where(name__invalid__op=0)
            str(q)

        # invalid ftm_columnstore_test props
        with self.assertRaisesRegex(InvalidQuery, "Invalid FtM property"):
            q = Query().where(invalid_prop=0)
            str(q)

        with self.assertRaisesRegex(InvalidQuery, "Invalid FtM property"):
            q = Query().where(invalid_prop__like=0)
            str(q)

    def test_entity_query(self):
        q = EntityQuery()
        self.assertEqual(
            str(q),
            "SELECT dataset, entity_id, schema, groupArray(prop) as props, groupArray(values) as values FROM (SELECT dataset, entity_id, schema, prop, groupUniqArray(value) as values FROM ftm_columnstore_test WHERE entity_id IN (SELECT DISTINCT entity_id FROM ftm_columnstore_test) GROUP BY dataset, entity_id, prop, schema) GROUP BY dataset, entity_id, schema ORDER BY dataset, entity_id, schema ASC",
        )

        # all filters etc. will be applied to the innerst query
        q = EntityQuery().where(entity_id=1)
        self.assertEqual(
            str(q),
            "SELECT dataset, entity_id, schema, groupArray(prop) as props, groupArray(values) as values FROM (SELECT dataset, entity_id, schema, prop, groupUniqArray(value) as values FROM ftm_columnstore_test WHERE entity_id IN (SELECT DISTINCT entity_id FROM ftm_columnstore_test WHERE entity_id = '1') GROUP BY dataset, entity_id, prop, schema) GROUP BY dataset, entity_id, schema ORDER BY dataset, entity_id, schema ASC",
        )

        q = EntityQuery()[:100]
        self.assertEqual(
            str(q),
            "SELECT dataset, entity_id, schema, groupArray(prop) as props, groupArray(values) as values FROM (SELECT dataset, entity_id, schema, prop, groupUniqArray(value) as values FROM ftm_columnstore_test WHERE entity_id IN (SELECT DISTINCT entity_id FROM ftm_columnstore_test LIMIT 0, 100) GROUP BY dataset, entity_id, prop, schema) GROUP BY dataset, entity_id, schema ORDER BY dataset, entity_id, schema ASC",
        )

        # make sure dataset is passed along
        q = EntityQuery().where(dataset="luanda_leaks")
        self.assertEqual(
            str(q),
            "SELECT dataset, entity_id, schema, groupArray(prop) as props, groupArray(values) as values FROM (SELECT dataset, entity_id, schema, prop, groupUniqArray(value) as values FROM ftm_columnstore_test WHERE dataset = 'luanda_leaks' AND entity_id IN (SELECT DISTINCT entity_id FROM ftm_columnstore_test WHERE dataset = 'luanda_leaks') GROUP BY dataset, entity_id, prop, schema) GROUP BY dataset, entity_id, schema ORDER BY dataset, entity_id, schema ASC",
        )



import logging

from .dataset import Dataset, get_dataset  # noqa
from .driver import get_driver  # noqa

# don't show clickhouse numpy warnings:
logging.getLogger("clickhouse_driver.columns.service").setLevel(logging.ERROR)



import uuid

from followthemoney import model
from tests.util import ClickhouseTestCase

from ftm_columnstore import Dataset


class IOTestCase(ClickhouseTestCase):
    def test_io_import(self):
        # testdata already imported during tearUp
        count = self.driver.execute(
            f"SELECT count(*) FROM {self.driver.table} WHERE dataset = 'luanda_leaks'"
        )
        count = count[0][0]
        self.assertEqual(count, 4342)

    def test_io_import_with_fpx(self):
        # testdata already imported during tearUp
        count = self.driver.execute(
            f"SELECT count(*) FROM {self.driver.table} WHERE dataset = 'luanda_leaks_fpx'"
        )
        count = count[0][0]
        self.assertEqual(count, 4342)

    def test_io_import_numeric(self):
        def _test_numeric_value(value, result_value):
            dataset = Dataset("test_numeric")
            entity = model.make_entity("Payment")
            entity.add("amount", value)
            entity.make_id(str(uuid.uuid4()))
            res = dataset.put(entity)
            self.assertEqual(res, 1)
            entity = dataset.get(entity.id)
            amount = entity.get("amount")[0]
            self.assertEqual(value, amount)  # ftm props always strings
            res = dataset.driver.query(
                f"SELECT * FROM {dataset.driver.table} WHERE entity_id = '{entity.id}'"
            )
            values = list(res)[0]
            value_num = values[8]
            if result_value is not None:
                self.assertIsInstance(value_num, float)
                self.assertEqual(round(value_num, 1), result_value)
            else:
                self.assertEqual(value_num, result_value)

        _test_numeric_value("1000", 1000)
        _test_numeric_value("1000.0", 1000)
        _test_numeric_value("1000.1", 1000.1)
        _test_numeric_value("1000,1", 1000.1)
        _test_numeric_value("1,000,000", 1_000_000)
        _test_numeric_value("1,000,000.1", 1_000_000.1)
        _test_numeric_value("1.000.000", 1_000_000)
        _test_numeric_value("1.000.000,1", 1_000_000.1)
        _test_numeric_value("1.000.000.1", None)
        _test_numeric_value("1,000,000,1", None)



import os


def get_env(name, default=None):
    value = os.environ.get(name)
    if value is not None:
        return str(value)
    if default is not None:
        return str(default)


DATABASE_URI = get_env("DATABASE_URI", "localhost")
DATABASE_TABLE = get_env("DATABASE_TABLE", "ftm")
LOG_LEVEL = get_env("LOG_LEVEL", "info")
LRU_QUERY_CACHE_SIZE = int(get_env("LRU_QUERY_CACHE_SIZE", 1024 * 1000))
BULK_WRITE_SIZE = int(get_env("BULK_WRITE_SIZE", 100_000))



import itertools
from typing import Any, Iterable, Iterator, Optional, Union

import pandas as pd
from banal import as_bool, clean_dict, is_listish
from followthemoney import model
from followthemoney.proxy import E

from . import enums
from .driver import ClickhouseDriver, get_driver
from .exceptions import InvalidQuery

TYPES = {p.name: p.type.name for p in model.properties}


class Query:
    OPERATORS = {
        "like": "LIKE",
        "ilike": "ILIKE",
        "gt": ">",
        "gte": ">=",
        "lt": "<",
        "lte": "<=",
        "in": "IN",
        "null": "IS",
    }
    fields = None

    def __init__(
        self,
        from_: Optional[Union[str, "Query"]] = None,
        fields: Optional[Iterable[str]] = None,
        group_by_fields: Optional[Iterable[str]] = None,
        order_by_fields: Optional[Iterable[str]] = None,
        order_direction: Optional[str] = "ASC",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        where_lookup: Optional[dict] = None,
        having_lookup: Optional[dict] = None,
        driver: Optional[ClickhouseDriver] = None,
    ):
        driver = driver or get_driver()
        if from_ is None:
            from_ = driver.table
        self.driver = driver
        self.fields = fields or self.fields
        self.from_ = f"({from_})" if isinstance(from_, Query) else from_
        self.group_by_fields = group_by_fields
        self.order_by_fields = order_by_fields
        self.order_direction = order_direction
        self.limit = limit
        self.offset = offset
        self.where_lookup = where_lookup
        self.having_lookup = having_lookup

    def __str__(self) -> str:
        return self.get_query()

    def __iter__(self) -> Iterator[Any]:
        return self.iterate()

    def __len__(self) -> int:
        return self.count()

    def iterate(self):
        return self.execute()

    def count(self):
        # FIXME this doesn't cover aggregated cases for `having`
        count_part = "*"
        if self.group_by_fields:
            count_part = f"DISTINCT {', '.join(self.group_by_fields)}"
        query = (
            f"SELECT count({count_part}) as count FROM {self.from_}{self.where_part}"
        )
        res = self.execute(query)
        for i in res:
            return i[0]

    def execute(self, query: Optional[Union["Query", str]] = None) -> Iterator[Any]:
        """return result iterator"""
        query = query or self.get_query()
        return self.driver.query(str(query))

    def first(self):
        # return the first object
        for res in self:
            return res

    def _chain(self, **kwargs):
        # merge current state
        new_kwargs = self.__dict__.copy()
        for key, new_value in kwargs.items():
            old_value = new_kwargs[key]
            if old_value is None:
                new_kwargs[key] = new_value
            # "remove" old value:
            elif new_value is None:
                new_kwargs[key] = None
            # overwrite order by
            elif key == "order_by_fields":
                new_kwargs[key] = new_value
            # combine iterables and dicts
            elif is_listish(old_value):
                new_kwargs[key] = sorted(set(old_value) | set(new_value))
            elif isinstance(old_value, dict):
                new_kwargs[key] = {**old_value, **new_value}
            else:  # replace
                new_kwargs[key] = new_value
        return self.__class__(**new_kwargs)

    def select(self, *fields) -> "Query":
        return self._chain(fields=fields)

    def where(self, **filters) -> "Query":
        return self._chain(where_lookup=filters)

    def having(self, **filters) -> "Query":
        return self._chain(having_lookup=filters)

    def group_by(self, *fields) -> "Query":
        return self._chain(group_by_fields=fields)

    def order_by(self, *fields, ascending=True) -> "Query":
        return self._chain(
            order_by_fields=fields, order_direction="ASC" if ascending else "DESC"
        )

    # for slicing
    def __getitem__(self, value) -> "Query":
        if isinstance(value, int):
            if value < 0:
                raise InvalidQuery("Invalid slicing: slice must not be negative.")
            return self._chain(limit=1, offset=value)
        if isinstance(value, slice):
            if value.step is not None:
                raise InvalidQuery("Invalid slicing: steps not allowed.")
            offset = value.start or 0
            if value.stop is not None:
                return self._chain(limit=value.stop - offset, offset=offset)
            return self._chain(offset=offset)
        raise NotImplementedError

    @property
    def fields_part(self) -> str:
        return ", ".join(self.fields or "*")

    def _get_lookup_part(
        self, lookup: dict, strict_fields: Optional[bool] = True
    ) -> str:
        """for where and having clause

        lookups for ftm properties (name="foo") will be rewritten as
        (prop="name" AND value="foo")
        """
        meta_fields = {
            "dataset",
            "schema",
            "origin",
            "entity_id",
            "prop",
            "prop_type",
            "value",
            "fingerprint",
            "fingerprint_id",
        }
        meta_parts = set()
        parts = set()
        lookup = clean_dict(lookup)

        def _get_part(key: str, value: str, operator: Optional[str] = None) -> str:
            operator = operator or "="
            if key in enums.PROPERTIES:
                value_field = "value"
                if TYPES.get(key) == "number":
                    value_field = "value_num"
                return f"(prop = '{key}' AND {value_field} {operator} {value})"
            return f"{key} {operator} {value}"

        for field, value in lookup.items():
            field, *operator = field.split("__")

            if strict_fields:
                if field not in meta_fields | enums.PROPERTIES:
                    if field not in enums.PROPERTIES:
                        raise InvalidQuery(f"Lookup `{field}`: Invalid FtM property.")
                    raise InvalidQuery(f"Lookup `{field}` not any of {meta_fields}")

            if operator:
                if len(operator) > 1:
                    raise InvalidQuery(f"Invalid operator: {operator}")
                operator = operator[0]
                if operator not in self.OPERATORS:
                    raise InvalidQuery(f"Invalid operator: {operator}")

                if operator == "in":
                    if isinstance(value, (str, Query)):
                        value = f"({value})"
                    elif is_listish(value) or isinstance(value, pd.Series):
                        values = ", ".join([f"'{v}'" for v in value])
                        value = f"({values})"
                    else:
                        raise InvalidQuery(f"Invalid value for `IN` operator: {value}")
                elif operator == "null":
                    # field__null=True|False
                    value = "NULL" if as_bool(value) else "NOT NULL"
                else:
                    value = f"'{value}'"
                part = _get_part(field, value, self.OPERATORS[operator])
            else:
                part = _get_part(field, f"'{value}'")

            if field in meta_fields:
                meta_parts.add(part)
            else:
                parts.add(part)

        final_parts = []
        if meta_parts:
            final_parts.append(
                " AND ".join(sorted(meta_parts))
            )  # sort for easier testing
        if parts:
            final_parts.append(" OR ".join(sorted(parts)))  # sort for easier testing
        return " AND ".join(final_parts)

    @property
    def where_part(self) -> str:
        if not self.where_lookup:
            return ""
        return " WHERE " + self._get_lookup_part(self.where_lookup)

    @property
    def having_part(self) -> str:
        if not self.group_part or not self.having_lookup:
            return ""
        return " HAVING " + self._get_lookup_part(self.having_lookup, False)

    @property
    def group_part(self) -> str:
        if self.group_by_fields is None:
            return ""
        return " GROUP BY " + ", ".join(sorted(self.group_by_fields))

    @property
    def order_part(self) -> str:
        if self.order_by_fields is None:
            return ""
        return (
            " ORDER BY " + ", ".join(self.order_by_fields) + " " + self.order_direction
        )

    @property
    def limit_part(self) -> str:
        if self.limit is None and self.offset is None:
            return ""
        offset = self.offset or 0
        if self.limit:
            if self.limit < 0:
                raise InvalidQuery(f"Limit {self.limit} must not be negative")
            return f" LIMIT {offset}, {self.limit}"
        return f" OFFSET {offset}"

    @property
    def is_filtered(self) -> bool:
        return bool(self.where_part or self.having_part)

    def get_query(self) -> str:
        return self.to_str(self)

    def as_table(self) -> pd.DataFrame:
        """pivot via pandas to have columns per prop"""
        df = self.driver.query_dataframe(str(self))
        df = (
            df.groupby(["dataset", "schema", "entity_id", "prop"])
            .agg({"value": lambda s: s.unique()})
            .reset_index()
        )
        df = df.pivot(["dataset", "entity_id", "schema"], "prop", "value").reset_index()
        return df

    @classmethod
    def to_str(cls, query: "Query") -> str:
        rest = "".join(
            (
                query.where_part,
                query.group_part,
                query.having_part,
                query.order_part,
                query.limit_part,
            )
        ).strip()
        q = f"SELECT {query.fields_part} FROM {query.from_} {rest}"
        return q.strip()


class EntityQuery(Query):
    """aggregate rows to entity instances and use where/having/order_by on
    actual properties"""

    fields = ("DISTINCT entity_id",)

    @property
    def dataset(self):
        if self.where_lookup is not None:
            return self.where_lookup.get("dataset")

    def __iter__(self) -> Iterator[E]:
        res = self.execute()
        entity = None
        for dataset, entity_id, schema, props, values in res:
            # result is already aggregated per (id, schema) and sorted via query,
            # so each row is 1 entity
            next_entity = model.get_proxy(
                {
                    "id": entity_id,
                    "schema": schema,
                    "properties": dict(zip(props, values)),
                }
            )
            if entity is None:
                entity = next_entity
            elif entity.id != next_entity.id:
                yield entity
                entity = next_entity
            else:
                entity.merge(next_entity)
        if entity is not None:
            yield entity

    def get_query(self) -> str:
        """
        first get matching entity_ids for given where clause(s),
        then return 1 row per dataset->entity_id->schema with props and values
        as arrays

        example:


        SELECT dataset, entity_id, schema, groupArray(prop) as props, groupArray(values) as values FROM (
            SELECT dataset, entity_id, schema, prop, groupUniqArray(value) as values FROM ftm
            WHERE entity_id IN (
                SELECT DISTINCT entity_id FROM ftm
                WHERE schema = 'Person' AND (prop = 'name' AND value = 'Simon')
            )
            GROUP BY dataset, entity_id, schema, prop
        )
        GROUP BY dataset, entity_id, schema
        ORDER BY dataset, entity_id, schema

        """
        inner = super().get_query()
        outer = (
            Query()
            .select(
                "dataset",
                "entity_id",
                "schema",
                "prop",
                "groupUniqArray(value) as values",
            )
            .where(entity_id__in=inner, dataset=self.dataset)
            .group_by("dataset", "entity_id", "schema", "prop")
        )
        return str(
            Query(outer)
            .select(
                "dataset",
                "entity_id",
                "schema",
                "groupArray(prop) as props",
                "groupArray(values) as values",
            )
            .group_by("dataset", "entity_id", "schema")
            .order_by(
                "dataset", "entity_id", "schema"
            )  # make sure ordering to easier build entities later
        )

    def iterate(self, chunksize: Optional[int] = 1000) -> Iterator[E]:
        # iterate in chunks, useful for huge bulk streaming performance
        q = self
        for i in itertools.count(0, chunksize):
            chunk = (x for x in q[i : i + chunksize])
            try:
                # maybe chunk is empty, then abort
                entity = next(chunk)
                yield entity
            except StopIteration:
                return

            while True:
                try:
                    entity = next(chunk)
                    yield entity
                except StopIteration:
                    break  # to next chunk

    def as_table(self):
        def _unpack(serie: pd.Series) -> pd.Series:
            return pd.Series([i for x in serie.values for i in x]).unique()

        inner = super().get_query()
        q = (
            Query()
            .select(
                "dataset",
                "entity_id",
                "schema",
                "prop",
                "groupUniqArray(value) as values",
            )
            .where(entity_id__in=inner)
            .group_by("dataset", "entity_id", "schema", "prop")
        )
        df = self.driver.query_dataframe(str(q))
        df = (
            df.groupby(["dataset", "schema", "entity_id", "prop"])
            .agg({"values": _unpack})
            .reset_index()
        )
        df = df.pivot(
            ["dataset", "entity_id", "schema"], "prop", "values"
        ).reset_index()
        return df



import os
from unittest import TestCase

from ftm_columnstore import settings
from ftm_columnstore.io import import_json
from ftm_columnstore.driver import get_driver


def get_clickhouse_test_driver():
    settings.DATABASE_URI = "localhost"
    settings.DATABASE_TABLE = "ftm_columnstore_test"
    driver = get_driver()
    assert driver.table == "ftm_columnstore_test"
    return driver


class ClickhouseTestCase(TestCase):
    data_file = os.path.join(
        os.getcwd(), "tests", "fixtures", "icij_luanda_leaks.jsonl"
    )

    @classmethod
    def setUpClass(cls):
        cls.driver = get_clickhouse_test_driver()
        cls.driver.init(recreate=True)
        res1 = import_json(cls.data_file, "luanda_leaks")
        res2 = import_json(cls.data_file, "luanda_leaks_fpx", with_fingerprints=True)
        assert res1 == 852  # number of entities
        assert res1 == res2

    @classmethod
    def tearDownClass(cls):
        cls.driver.dangerous_drop()



from tests.util import ClickhouseTestCase

from ftm_columnstore.dataset import Dataset


class DatasetTestCase(ClickhouseTestCase):
    def test_dataset_init(self):
        ds = Dataset("luanda_leaks")
        self.assertEqual(ds.driver.table, "ftm_columnstore_test")
        self.assertEqual(ds.driver.table_fpx, "ftm_columnstore_test_fpx")
        self.assertIsNone(ds.origin)

    def test_dataset_iteration(self):
        ds = Dataset("luanda_leaks")
        entities = [e for e in ds]
        self.assertEqual(len(entities), 852)
        entities = [e for e in ds.iterate(chunksize=500)]
        self.assertEqual(len(entities), 852)

    def test_dataset_statements_iteration(self):
        ds = Dataset("luanda_leaks")
        stmts = [s for s in ds.statements()]
        self.assertEqual(len(stmts), 4342)

    def test_dataset_get_entity(self):
        ds = Dataset("luanda_leaks")
        entity = ds.get("0372a4b5d9c3f01f5b9eb1dddf8677ecc777b0a3")
        self.assertDictEqual(
            entity.to_dict(),
            {
                "id": "0372a4b5d9c3f01f5b9eb1dddf8677ecc777b0a3",
                "schema": "Company",
                "properties": {
                    "alephUrl": [
                        "https://aleph.occrp.org/api/2/entities/0372a4b5d9c3f01f5b9eb1dddf8677ecc777b0a3.f14091b5d76ff7b3b2c846ba4ce68395ac36145c"
                    ],
                    "sector": ["Oil, Gas and Coal"],
                    "dissolutionDate": ["2014-12-30"],
                    "name": ["Sopor - Sociedade Distribuidora de Combustíveis, S.A."],
                    "incorporationDate": ["1957-07-17"],
                    "proof": ["6ea1f60f2634552390f3e9d1ac1b81a119d6d73e"],
                    "jurisdiction": ["pt"],
                },
            },
        )

    def test_dataset_entity_resolve(self):
        ds = Dataset("luanda_leaks")
        entity = ds.get("003e86f3eae13d711ce9771f6bc3e47142069a7d")  # Ownership entity
        entities = [e for e in ds.resolve(entity)]
        self.assertEqual(2, len(entities))

    def test_dataset_entity_expand(self):
        ds = Dataset("luanda_leaks")
        isabel = ds.get("dfa01f04a87a4ce0f1e0f96a85d260a3249ab64a")
        entities = [e for e in ds.expand(isabel)]
        self.assertEqual(len(entities), 109)
        self.assertSetEqual(
            set([e.schema.name for e in entities]), set(["Ownership", "Table"])
        )

        # more levels
        entities = [e for e in ds.expand(isabel, levels=2)]
        self.assertEqual(len(entities), 219)



from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()


setup(
    name="ftm-columnstore",
    version="0.0.3",
    description="Column store implementation for ftm data based on clickhouse",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Simon Wörpel",
    author_email="simon.woerpel@medienrevolte.de",
    url="https://github.com/simonwoerpel/ftm-columnstore",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(),
    package_dir={"ftm_columnstore": "ftm_columnstore"},
    install_requires=[
        "banal",
        "Click",
        "clickhouse-driver[numpy]",
        "followthemoney",
        "pandas",
        "pyicu",
        "structlog",
    ],
    entry_points={"followthemoney.cli": ["cstore = ftm_columnstore.cli:cli"]},
)



import csv
import json
import logging
from typing import Optional

import click
from followthemoney.cli.cli import cli as main
from followthemoney.cli.util import MAX_LINE, write_object

from . import settings, statements
from .dataset import Dataset
from .driver import get_driver

log = logging.getLogger(__name__)


def readlines(stream):
    while True:
        line = stream.readline(MAX_LINE)
        if not line:
            return
        yield line.strip()


def _get_driver(obj):
    return get_driver(uri=obj["uri"], table=obj["table"])


def _get_dataset(
    obj,
    name: str,
    origin: Optional[str] = None,
    ignore_errors: Optional[bool] = False,
):
    driver = _get_driver(obj)
    return Dataset(name, origin=origin, driver=driver, ignore_errors=ignore_errors)


@click.group(help="Store FollowTheMoney object data in a column store (Clickhouse)")
@click.option(
    "--log-level",
    default=settings.LOG_LEVEL,
    help="Set logging level",
    show_default=True,
)
@click.option(
    "--uri",
    help="Database connection URI",
    default=settings.DATABASE_URI,
    show_default=True,
)
@click.option(
    "--table",
    help="Database table",
    default=settings.DATABASE_TABLE,
    show_default=True,
)
@click.pass_context
def cli(ctx, log_level, uri, table):
    """
    FtM Columnstore implementation (Clickhouse)
    """
    logging.basicConfig(level=log_level)
    ctx.obj = {"uri": uri, "table": table}
    log.info(f"Using database driver: `{uri}` (table: `{table}`)")


@cli.command("flatten")
@click.option("-i", "--infile", type=click.File("r"), default="-")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.option("-d", "--dataset", help="Dataset", required=True)
@click.option(
    "--header/--no-header",
    type=bool,
    default=True,
    help="Print header row or not",
    show_default=True,
)
def flatten(infile, outfile, dataset, header):
    """Turn json entities from `infile` into statements in csv format"""
    writer = csv.DictWriter(outfile, fieldnames=statements.COLUMNS)
    if header:
        writer.writeheader()
    for data in readlines(infile):
        data = json.loads(data)
        for stmt in statements.statements_from_entity(data, dataset):
            writer.writerow(stmt)


@cli.command("fingerprints")
@click.option("-i", "--infile", type=click.File("r"), default="-")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.option("-d", "--dataset", help="Dataset", required=True)
@click.option(
    "--header/--no-header",
    type=bool,
    default=True,
    help="Print header row or not",
    show_default=True,
)
def flatten_fingerprints(infile, outfile, dataset, header):
    """Generate fingerprint statements as csv from json entities from `infile`"""
    writer = csv.DictWriter(outfile, fieldnames=statements.COLUMNS_FPX)
    if header:
        writer.writeheader()
    for data in readlines(infile):
        data = json.loads(data)
        for stmt in statements.fingerprints_from_entity(data, dataset):
            writer.writerow(stmt)


@cli.command("init", help="Initialize database and table.")
@click.option(
    "--recreate/--no-recreate",
    help="Recreate database if it already exists.",
    default=False,
    show_default=True,
)
@click.pass_obj
def db_init(obj, recreate):
    driver = _get_driver(obj)
    driver.init(recreate=recreate)


@cli.command("write")
@click.option("-i", "--infile", type=click.File("r"), default="-")
@click.option("-d", "--dataset", help="dataset identifier", required=True)
@click.option("-o", "--origin", default="bulk")
@click.option(
    "--fingerprints/--no-fingerprints",
    help="Populate fingerprints ix table",
    default=False,
    show_default=True,
    type=bool,
)
@click.option(
    "--ignore-errors/--no-ignore-errors",
    type=bool,
    default=False,
    help="Don't fail on errors, only log them.",
    show_default=True,
)
@click.pass_obj
def write(obj, infile, dataset, origin, fingerprints, ignore_errors):
    """
    Write json entities from `infile` to store.
    """
    dataset = _get_dataset(
        obj,
        dataset,
        origin=origin,
        ignore_errors=ignore_errors,
    )

    bulk = dataset.bulk(with_fingerprints=fingerprints)
    for line in readlines(infile):
        entity = json.loads(line)
        bulk.put(entity)
    bulk.flush()


@cli.command("iterate", help="Iterate entities")
@click.option("-d", "--dataset", help="Dataset")
@click.option("--origin")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.pass_obj
def iterate(obj, dataset, origin, outfile):
    dataset = _get_dataset(obj, dataset, origin)
    for entity in dataset.iterate():
        write_object(outfile, entity)


@cli.command("statements", help="Dump all statements as csv")
@click.option("-d", "--dataset", help="Dataset")
@click.option("--origin")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.pass_obj
def dump_statements(obj, dataset, origin, outfile):
    dataset = _get_dataset(obj, dataset, origin)
    writer = csv.writer(outfile)
    writer.writerow(statements.COLUMNS)
    for row in dataset.statements(origin=origin):
        writer.writerow(row)


@cli.command("delete", help="Delete dataset or complete store")
@click.option("-d", "--dataset", help="Dataset")
@click.option("-o", "--origin")
@click.pass_obj
def delete(obj, dataset, origin):
    dataset = _get_dataset(obj, dataset, origin)
    dataset.delete(origin=origin)


@cli.command("list", help="List datasets in a store")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.option(
    "--verbose",
    help="Include entity count numbers",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_obj
def list_datasets(obj, outfile, verbose):
    driver = _get_driver(obj)
    q = f"SELECT distinct dataset FROM {driver.table}"
    if verbose:
        q = f"""SELECT dataset,
        count(distinct entity_id) as entities,
        count(*) as statements
        FROM {driver.table} GROUP BY dataset"""
    df = driver.query_dataframe(q)
    df.to_csv(outfile, index=False)


@cli.command("query", help="Execute raw query and print result (csv format) to outfile")
@click.argument("query")
@click.option("-o", "--outfile", type=click.File("w"), default="-")
@click.pass_obj
def db_query(obj, query, outfile):
    driver = _get_driver(obj)
    df = driver.query_dataframe(query)
    df.to_csv(outfile, index=False)


# Register with main FtM command-line tool.
main.add_command(cli, name="cstore")



from followthemoney import model

SCHEMATA = set(model.schemata.keys())
PROPERTIES = set([p.name for p in model.properties])
PROPERTIES_FPX = set([p.name for p in model.properties if p.type.name == "name"])
PROPERTY_TYPES = set(p.type.name for p in model.properties)



# shorthands for import / export

import json
from typing import Optional

from followthemoney import model

from ftm_columnstore.dataset import Dataset
from ftm_columnstore.driver import ClickhouseDriver, get_driver


def import_json(
    fpath: str,
    dataset: str,
    with_fingerprints: Optional[bool] = False,
    origin: Optional[str] = None,
    driver: Optional[ClickhouseDriver] = None,
    **kwargs
) -> int:
    """
    import a json file containing ftm entities, 1 entity per line without comma
    separator (aka jsonlines)

    return: number of imported entities
    """
    driver = driver or get_driver()
    dataset = Dataset(dataset, origin, driver, **kwargs)
    bulk = dataset.bulk(with_fingerprints=with_fingerprints)
    i = 0
    with open(fpath) as f:
        for line in f.readlines():
            entity = json.loads(line)
            entity = model.get_proxy(entity)  # validation
            bulk.put(entity)
            i += 1
    bulk.flush()
    return i
